#define ROWS 4
#define COLS 4
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <omp.h>


// für sicherung / Kommunikation
typedef signed char uchar;
typedef uchar field[ROWS*COLS+1]; //das erste Element encodiert die momentane Tiefe des Feldes (wie viele Züge gespielt wurden), die restlichen geben die Einwurfspalte an
typedef uchar field_xy[ROWS*COLS]; //konvertiertes 2D-Feld um member mit acces_xy zu erreichen
#define access_xy(c, x, y) c[x + COLS *(y)]
#define CHUNK_SIZE 1

// Globale Variablen
uchar throwable;
uchar board_full;
uchar winner_exs;
uchar game_finished;
int zaehler;
uchar current_player;
uchar comparison[COLS];
uchar help_comparison[COLS];
uchar best_turn;

// checks for free space in first row, returns 1 if there is space
static uchar can_throw(field_xy c, uchar row){
   if(access_xy(c,row,ROWS-1) == 0) {
    return 1; //oberste Reihe der Spalte leer
   }
   else {
    return 0; //Oberste Reihe der Spalte ist voll
   }
//access_xy(f,row,COLS-1) == 0; // Verfügbarer Platz
}

//Checks if the game is finished and writes it into a global variable, 1 = there is a winner
static uchar is_winner(field_xy f){
    uchar i = 0;
    uchar j = 0;
    board_full = 0;
    for(j=0;j<ROWS;j++) { //horizontales iterieren
        for(i=0;i<COLS-3;i++) {
            if((access_xy(f, i, j) != 0) && (access_xy(f, i, j) == access_xy(f, i+1, j)) && (access_xy(f, i, j) == access_xy(f, i+2, j)) && (access_xy(f, i, j) == access_xy(f, i+3, j))) {
                winner_exs = 1;
                goto end;
            }
            else {
                winner_exs = 0;
            }
        }
    }
    for(i=0;i<COLS;i++) { //vertikales iterieren
        for(j=0;j<ROWS-3;j++) {
            if((access_xy(f, i, j) != 0) && (access_xy(f, i, j) == access_xy(f, i, j+1)) && (access_xy(f, i, j) == access_xy(f, i, j+2)) && (access_xy(f, i, j) == access_xy(f, i, j+3))) {
                winner_exs = 1;
                goto end;
            }
            else {
                winner_exs = 0;
            }
        }
    }
    for(i=0;i<COLS-3;i++) { //rechts oben iterieren
        for(j=0;j<ROWS-3;j++) {
            if((access_xy(f, i, j) != 0) && (access_xy(f, i, j) == access_xy(f, i+1, j+1)) && (access_xy(f, i, j) == access_xy(f, i+2, j+2)) && (access_xy(f, i, j) == access_xy(f, i+3, j+3))) {
                winner_exs = 1;
                goto end;
            }
            else {
                winner_exs = 0;
            }
        }
    }
    for(i=3;i<COLS;i++) { //links oben iterieren
        for(j=0;j<ROWS-3;j++) {
            if((access_xy(f, i, j) != 0) && (access_xy(f, i, j) == access_xy(f, i-1, j+1)) && (access_xy(f, i, j) == access_xy(f, i-2, j+2)) && (access_xy(f, i, j) == access_xy(f, i-3, j+3))) {
                winner_exs = 1;
                goto end;
            }
            else {
                winner_exs = 0;
            }
        }
    }

//Bei Spielende =1 nicht bei Spielende =0
    end: return winner_exs;
}

static void init_comp() {
    uchar i;
    for(i=0;i<COLS;i++) {
        comparison[i] = -2 ;
        help_comparison[i] = -2;
    }
}

//Checks if board is full, returns 0 if board is not full
static uchar is_board_full(field f) {
    if (f[0] == (COLS*ROWS)) {
        return 1;
    }
    else {
        return 0;
    }
}

//prints the field
static void print_field(field_xy c){
  printf("\n");
  for(int y=ROWS-1; y >= 0; y--){
	for(int x=0; x < COLS; x++){
		printf("%d ", access_xy(c, x, y));
	}
	printf("\n");
  }
}

//copy gamestate
static void copy_field(field f, field v) {
    int i = 0;
    for(i=0;i<=(COLS*ROWS);i++) {
        v[i] = f[i];
    }
}

//performs a throw
static void throw(field f, uchar row){
	f[f[0]] = row;
	f[0] = f[0] + 1;
//	assert(f[0] <= (ROWS*COLS+1));
}

//converts the gamestate-format
static void convert_1D_to_2D(field f, field_xy out){
	uchar depth[COLS]; // momentane Tiefe
	memset(depth, 0, COLS);
	memset(out, 0, ROWS*COLS);
	for(uchar i=1; i < f[0]; i++){
		int dep = depth[f[i]]++;
		access_xy(out,f[i], dep) = i % 2 + 1;
	}
}

//initializes the game
static void init_field(field f){
	f[0] = 1;
}

//checks whos turn it is
static void whos_turn(field f) {
    if(f[0] % 2 == 1) {
        current_player = 1;
    }
    else if(f[0] % 2 == 0) {
        current_player = 2;
    }
}

//undo a turn in gamestate
static void undo_turn(field g) {
    g[g[0] - 1] = 0 ;
    g[0]--;
}

//evaluates a turn
static uchar evaluate_turn(field g) {
    field_xy c;
    convert_1D_to_2D(g, c);
    if (((g[0] % 2) == 1) && (current_player == 1) && (is_winner(c) == 1) ) { //Spieler 1 hat gewonnen und für ihn wird gezählt || +1
        return 1;
    }
    else if (((g[0] % 2) == 0) && (current_player == 2) && (is_winner(c) == 1) ) { //Spieler 2 hat gewonnen und für ihn wird gezählt || +1
        return 1;
    }
    else if (((g[0] % 2) == 1) && (current_player == 2) && (is_winner(c) == 1) ) { //Spieler 2 hat gewonnen und für 1 wird gezählt || -1
        return -1;
    }
    else if (((g[0] % 2) == 0) && (current_player == 1) && (is_winner(c) == 1) ) { //Spieler 1 hat gewonnen und für 2 wird gezählt || -1
        return -1;
    }
    else if ((is_board_full(g) == 1) && (is_winner(c) == 0)) { //Spielfeld voll , kein Sieger - unentschieden || 0
        return 0;
    }
}

//checking possible throws
void minimax(field f, field g){
    uchar p;
    field_xy c;
    convert_1D_to_2D(g, c);
    for(p=0;p<COLS;p++) {
        if(can_throw(c, p) == 1) {
            throw(g, p);
            convert_1D_to_2D(g, c);
            if((is_winner(c) == 0) && (is_board_full(g) == 0)) { //es gibt keinen Sieger und das Spielfeld is noch frei
                minimax(f, g);
                undo_turn(g);
            }
            else if(is_winner(c) == 1) { //es gibt einen Sieger
                help_comparison[g[f[0]]] = evaluate_turn(g);
                if ((current_player == 1) && (g[0] % 2 == 1) && (help_comparison[g[f[0]]] < comparison[g[f[0]]])) {
                    comparison[g[f[0]]] = help_comparison[g[f[0]]];
                }
                else if ((current_player == 1) && (g[0] % 2 == 0) && (help_comparison[g[f[0]]] > comparison[g[f[0]]])) {
                    comparison[g[f[0]]] = help_comparison[g[f[0]]];
                }
                else if ((current_player == 2) && (g[0] % 2 == 1) && (help_comparison[g[f[0]]] > comparison[g[f[0]]])) {
                    comparison[g[f[0]]] = help_comparison[g[f[0]]];
                }
                else if ((current_player == 2) && (g[0] % 2 == 0) && (help_comparison[g[f[0]]] < comparison[g[f[0]]])) {
                    comparison[g[f[0]]] = help_comparison[g[f[0]]];
                }
                undo_turn(g);
            }
            else if(is_winner(c) == 0 && (is_board_full(g) == 1)) { //es gibt keinen Sieger aber das Spielfeld ist voll
                help_comparison[g[f[0]]] = evaluate_turn(g);
                if ((current_player == 1) && (g[0] % 2 == 1) && (help_comparison[g[f[0]]] < comparison[g[f[0]]])) {
                    comparison[g[f[0]]] = help_comparison[g[f[0]]];
                }
                else if ((current_player == 1) && (g[0] % 2 == 0) && (help_comparison[g[f[0]]] > comparison[g[f[0]]])) {
                    comparison[g[f[0]]] = help_comparison[g[f[0]]];
                }
                else if ((current_player == 2) && (g[0] % 2 == 1) && (help_comparison[g[f[0]]] > comparison[g[f[0]]])) {
                    comparison[g[f[0]]] = help_comparison[g[f[0]]];
                }
                else if ((current_player == 2) && (g[0] % 2 == 0) && (help_comparison[g[f[0]]] < comparison[g[f[0]]])) {
                    comparison[g[f[0]]] = help_comparison[g[f[0]]];
                }
                undo_turn(g);
            }
        }
    }
}
//same as minimax() but for first level of tree
static void minimax_seq(field f, field g) {
    uchar i;
    uchar p;
    field_xy c;
    convert_1D_to_2D(g, c);
    #pragma omp parallel
    {
    #pragma omp single
    {
    for(i=0;i<COLS;i++) {
        #pragma omp task firstprivate(g, c, i, p)
        {
        if(can_throw(c, i) == 1) {
            throw(g, i);
            convert_1D_to_2D(g, c);
            if((is_winner(c) == 0) && (is_board_full(g) == 0)) { //es gibt keinen Sieger und das Spielfeld is noch frei
                minimax(f, g);
                undo_turn(g);
            }
            else if(is_winner(c) == 1) { //es gibt einen Sieger
                help_comparison[g[f[0]]] = evaluate_turn(g);
                if ((current_player == 1) && (g[0] % 2 == 1) && (help_comparison[g[f[0]]] < comparison[g[f[0]]])) {
                    comparison[g[f[0]]] = help_comparison[g[f[0]]];
                }
                else if ((current_player == 1) && (g[0] % 2 == 0) && (help_comparison[g[f[0]]] > comparison[g[f[0]]])) {
                    comparison[g[f[0]]] = help_comparison[g[f[0]]];
                }
                else if ((current_player == 2) && (g[0] % 2 == 1) && (help_comparison[g[f[0]]] > comparison[g[f[0]]])) {
                    comparison[g[f[0]]] = help_comparison[g[f[0]]];
                }
                else if ((current_player == 2) && (g[0] % 2 == 0) && (help_comparison[g[f[0]]] < comparison[g[f[0]]])) {
                    comparison[g[f[0]]] = help_comparison[g[f[0]]];
                }
                undo_turn(g);
            }
            else if(is_winner(c) == 0 && (is_board_full(g) == 1)) { //es gibt keinen Sieger aber das Spielfeld ist voll
                help_comparison[g[f[0]]] = evaluate_turn(g);
                if ((current_player == 1) && (g[0] % 2 == 1) && (help_comparison[g[f[0]]] < comparison[g[f[0]]])) {
                    comparison[g[f[0]]] = help_comparison[g[f[0]]];
                }
                else if ((current_player == 1) && (g[0] % 2 == 0) && (help_comparison[g[f[0]]] > comparison[g[f[0]]])) {
                    comparison[g[f[0]]] = help_comparison[g[f[0]]];
                }
                else if ((current_player == 2) && (g[0] % 2 == 1) && (help_comparison[g[f[0]]] > comparison[g[f[0]]])) {
                    comparison[g[f[0]]] = help_comparison[g[f[0]]];
                }
                else if ((current_player == 2) && (g[0] % 2 == 0) && (help_comparison[g[f[0]]] < comparison[g[f[0]]])) {
                    comparison[g[f[0]]] = help_comparison[g[f[0]]];
                }
                undo_turn(g);
            }
        }
    }}}
    #pragma omp taskwait
    }
}

//evaluates the best turn from the comparison-array
static void get_turn(field f) {
    field_xy c;
    convert_1D_to_2D(f, c);
    uchar i;
    best_turn = -2 ;
    for (i=0;i<COLS;i++) {
        if(can_throw(c, i) == 0) {
            comparison[i] = -3 ;
        }
        if (comparison[i] > best_turn) {
            best_turn = i ;
        }
    }
}

//checks if the game is over
static void is_game_over(field_xy c) {
    if((is_winner(c) == 1) || (is_board_full(c) == 1)) {
            game_finished = 1;
        }
        else {
            game_finished = 0;
        }
}

//a full turn
static void ai_turn(field f, field g) {
    field_xy c;
    copy_field(f, g);
    whos_turn(f);
    {
        minimax_seq(f, g);
    }
    get_turn(f);
    throw(f, best_turn);
    convert_1D_to_2D(f, c);
    is_game_over(c);
    print_field(c);
    printf("\n");
}

int main() {
    omp_set_num_threads(1);
    static field f;
    static field g;
    static field_xy c;
    init_field(f);
    init_comp();
    while(game_finished == 0) {
            ai_turn(f, g);
    }
    convert_1D_to_2D(f, c);
    is_game_over(c);
    print_field(c);
    printf("game finished: %d\n", game_finished);
    return 0;
}
